# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "pyramid_swagger"
__summary__ = "Swagger tools for use in pyramid webapps"
__uri__ = "https://github.com/striglia/pyramid_swagger"

__version__ = "2.3.0-rc2"

__author__ = "Scott Triglia"
__email__ = "scott.triglia@gmail.com"

__license__ = "BSD 3-clause"
__copyright__ = "Copyright 2014 Scott Triglia"
